# a modified version of aiwolf-python's client.py that can pass raw json to player
# Please note that Player class imported here is different from that in aiwolf-python,
# this class accepts more primitive implementation
import json
import socket
from typing import List, Optional, TypedDict
from aiwolf.gameinfo import _GameInfo
from aiwolf.gamesetting import _GameSetting
from aiwolfpy.player import AbstractPlayer
from aiwolf.utterance import _Utterance


class _Packet(TypedDict):
    gameInfo: Optional[_GameInfo]
    gameSetting: Optional[_GameSetting]
    request: str
    talkHistory: Optional[List[_Utterance]]
    whisperHistory: Optional[List[_Utterance]]


class TcpipClient:
    """Client player that communicates with the server via TCP/IP connection."""

    player: AbstractPlayer
    name: Optional[str]
    host: str
    port: int
    request_role: str
    sock: Optional[socket.socket]

    def __init__(self, player: AbstractPlayer, name: Optional[str], host: str, port: int, request_role: str) -> None:
        """Initialize a new instance of TcpipClient.

        Args:
            player: An AbstractPlayer to be connect with the server.
            name: The name of the player player.
            host: The hostname of the server.
            port: The port number the server is waiting on.
            request_role: The name of role that the player player wants to be.
        """
        self.player = player
        self.name = name
        self.host = host
        self.port = port
        self.request_role = request_role

    def _send_response(self, response: Optional[str]) -> None:
        if isinstance(self.sock, socket.socket) and isinstance(response, str):
            self.sock.send((response + "\n").encode("utf-8"))

    def _get_response(self, packet: _Packet) -> Optional[str]:
        request: str = packet["request"]
        game_info: Optional[_GameInfo] = packet["gameInfo"]
        game_setting: Optional[_GameSetting] = packet["gameSetting"]
        talk_history: Optional[List[_Utterance]] = packet["talkHistory"]
        whisper_history: Optional[List[_Utterance]] = packet["whisperHistory"]

        if request == "NAME":
            return self.name if self.name is not None else self.player.get_name()
        elif request == "ROLE":
            return self.request_role

        if request == "INITIALIZE":
            self.player.initialize(game_info, game_setting)
            return None
        else:
            self.player.update(game_info, talk_history, whisper_history, request)
            if request == "DAILY_INITIALIZE":
                self.player.day_start()
                return None
            elif request == "DAILY_FINISH":
                return None
            elif request == "FINISH":
                self.player.finish()
                return None
            elif request == 'VOTE':
                return json.dumps({'agentIdx': int(self.player.vote())}, separators=(',', ':'))
            elif request == 'ATTACK':
                return json.dumps({'agentIdx': int(self.player.attack())}, separators=(',', ':'))
            elif request == 'GUARD':
                return json.dumps({'agentIdx': int(self.player.guard())}, separators=(',', ':'))
            elif request == 'DIVINE':
                return json.dumps({'agentIdx': int(self.player.divine())}, separators=(',', ':'))
            elif request == 'TALK':
                return self.player.talk().__str__()
            elif request == 'WHISPER':
                return self.player.whisper().__str__()

            return None

    def connect(self) -> None:
        """Connect to the server."""
        self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.sock.settimeout(0.001)
        self.sock.connect((self.host, self.port))
        line: str = ""
        while True:
            try:
                line += self.sock.recv(8192).decode("utf-8")
                if line == "":
                    break
            except socket.timeout:
                pass
            line_list: List[str] = line.split("\n", 1)
            for i in range(len(line_list) - 1):
                if len(line_list[i]) > 0:
                    self._send_response(self._get_response(json.loads(line_list[i])))
                line = line_list[-1]
            try:
                self._send_response(self._get_response(json.loads(line)))
                line = ""
            except ValueError:
                pass
        self.sock.close()
        return None
