from itertools import combinations
from collections import deque
import time
import numpy as np

from aiwolf import GameInfo, GameSetting, Talk, Whisper, Content, Topic, Role, Species
from aiwolfpy.model.param_05 import param_05


village_size = 5
case_size = 20


# 0: VILLAGERS
# 1: POSSESSED
# 2: WEREWOLVES
case_mat = np.zeros((case_size, village_size), dtype=np.int32)
for _i, _c in enumerate(combinations(np.arange(village_size), 2)):
    for _p in range(2):
        _j = _i * 2 + _p
        case_mat[_j, _c] = 2
        case_mat[_j, _c[_p]] = 1

mat_vv = np.zeros((village_size, village_size, case_size), dtype=np.int32)
mat_vp = np.zeros((village_size, village_size, case_size), dtype=np.int32)
mat_vw = np.zeros((village_size, village_size, case_size), dtype=np.int32)
mat_pv = np.zeros((village_size, village_size, case_size), dtype=np.int32)
mat_pw = np.zeros((village_size, village_size, case_size), dtype=np.int32)
mat_wv = np.zeros((village_size, village_size, case_size), dtype=np.int32)
mat_wp = np.zeros((village_size, village_size, case_size), dtype=np.int32)
mat_ww = np.zeros((village_size, village_size, case_size), dtype=np.int32)
for _i in range(village_size):
    for _j in range(village_size):
        mat_vv[_i, _j, :] = (case_mat[:, _i] == 0) * (case_mat[:, _j] == 0)
        mat_vp[_i, _j, :] = (case_mat[:, _i] == 0) * (case_mat[:, _j] == 1)
        mat_vw[_i, _j, :] = (case_mat[:, _i] == 0) * (case_mat[:, _j] == 2)
        mat_pv[_i, _j, :] = (case_mat[:, _i] == 1) * (case_mat[:, _j] == 0)
        mat_pw[_i, _j, :] = (case_mat[:, _i] == 1) * (case_mat[:, _j] == 2)
        mat_wv[_i, _j, :] = (case_mat[:, _i] == 2) * (case_mat[:, _j] == 0)
        mat_wp[_i, _j, :] = (case_mat[:, _i] == 2) * (case_mat[:, _j] == 1)
        mat_ww[_i, _j, :] = (case_mat[:, _i] == 2) * (case_mat[:, _j] == 2)

base_prob = np.zeros(8, dtype=np.float32)
base_prob[0] = mat_vv[0, 1, :].sum() / case_size
base_prob[1] = mat_vp[0, 1, :].sum() / case_size
base_prob[2] = mat_vw[0, 1, :].sum() / case_size
base_prob[3] = mat_pv[0, 1, :].sum() / case_size
base_prob[4] = mat_pw[0, 1, :].sum() / case_size
base_prob[5] = mat_wv[0, 1, :].sum() / case_size
base_prob[6] = mat_wp[0, 1, :].sum() / case_size
base_prob[7] = mat_ww[0, 1, :].sum() / case_size


class RoleEstimator05:

    # pair_features
    # 0: vote
    # 1: declare vote
    # 2: divine werewolf
    # 3: divine human
    # 4: identify werewolf
    # 5: identify human
    # 6: guarded

    # single_features
    # 0: no co
    # 1: co seer
    # 2: co medium
    # 3: co bodyguard
    # 4: co villager
    # 5: co possessed
    # 6: co werewolf
    # 7: attacked

    def __init__(self):
        self.n_single_features = 8
        self.n_pair_features = 7
        # features
        self.single_features = np.zeros((village_size, self.n_single_features), dtype=np.float32)
        self.pair_features = np.zeros((village_size, village_size, self.n_pair_features), dtype=np.float32)
        self.co = np.zeros(village_size, dtype=np.int32)
        self.vote_declare = np.zeros(village_size, dtype=np.int32)
        self.vote_declare -= 1
        self.status = np.zeros(village_size, dtype=np.int32)
        # controls
        self.day = 0
        self.talk_idx = 0
        self.whisper_idx = 0
        # odds
        self.log_odds_list = np.ones(case_size, dtype=np.float32)
        self.single_log_odds = np.zeros((village_size, 3), dtype=np.float32)
        self.pair_log_odds = np.zeros((village_size, village_size, 8), dtype=np.float32)
        self.update_queue = deque()
        # agent_prob
        self.agent_prob = np.ones((village_size, 15), dtype=np.float32) / 15
        self.agent_param = np.zeros((village_size, param_05.shape[1], 2), dtype=np.float32)
        for i in range(village_size):
            for j in range(15):
                self.agent_param[i, :, :] += self.agent_prob[i, j] * param_05[j, :, :]

    def initialize(self, agent_prob=None):
        # features
        self.single_features *= 0.0
        self.pair_features *= 0.0
        self.co *= 0
        self.vote_declare *= 0
        self.vote_declare -= 1
        self.status *= 0
        # controls
        self.day = 0
        self.talk_idx = 0
        self.whisper_idx = 0
        # prob
        self.log_odds_list *= 0.0
        self.single_log_odds *= 0.0
        self.pair_log_odds *= 0.0
        self.update_queue = deque()
        # agent_prob
        if agent_prob is None:
            self.agent_prob = np.ones((village_size, 15), dtype=np.float32) / 15
        elif agent_prob.sum(axis=1).min() <= 0.0001:
            self.agent_prob = np.ones((village_size, 15), dtype=np.float32) / 15
        else:
            self.agent_prob = agent_prob / agent_prob.sum(axis=1, keepdims=True)

        self.agent_param = np.zeros((village_size, param_05.shape[1], 2), dtype=np.float32)
        for i in range(village_size):
            for j in range(15):
                self.agent_param[i, :, :] += self.agent_prob[i, j] * param_05[j, :, :]

    def update(self, game_info: GameInfo):
        t0 = time.time()
        # day start
        if game_info.day > self.day:
            # read vote_list
            for vote in game_info.vote_list:
                self.pair_features[
                    (vote.agent.agent_idx - 1) % village_size, (vote.target.agent_idx - 1) % village_size, 0
                ] = 1
                self.update_queue.append(
                    ((vote.agent.agent_idx - 1) % village_size, (vote.target.agent_idx - 1) % village_size)
                )
            # find executed agent
            if game_info.executed_agent is not None:
                self.status[(game_info.executed_agent.agent_idx - 1) % village_size] = 2
            # find attacked agent
            for agent in game_info.last_dead_agent_list:
                if self.status[(agent.agent_idx - 1) % village_size] == 0:
                    self.status[(agent.agent_idx - 1) % village_size] = 1
                    self.single_features[(agent.agent_idx - 1) % village_size, 7] = 1
                    self.update_queue.append(
                        ((agent.agent_idx - 1) % village_size, (agent.agent_idx - 1) % village_size)
                    )
            # refresh
            self.talk_idx = 0
            self.whisper_idx = 0
            for i in range(village_size):
                if self.vote_declare[i] != -1:
                    self.update_queue.append((i, self.vote_declare[i]))
            self.vote_declare *= 0
            self.vote_declare -= 1
            # update
            self.day = game_info.day

        # read talk
        if len(game_info.talk_list) > self.talk_idx:
            tk: Talk
            for tk in game_info.talk_list[self.talk_idx:]:

                content: Content = Content.compile(tk.text)

                # CO
                if content.topic == Topic.COMINGOUT:
                    if content.role == Role.SEER:
                        self.co[(tk.agent.agent_idx - 1) % village_size] = 1
                    elif content.role == Role.POSSESSED:
                        self.co[(tk.agent.agent_idx - 1) % village_size] = 2
                    elif content.role == Role.WEREWOLF:
                        self.co[(tk.agent.agent_idx - 1) % village_size] = 3
                    self.update_queue.append(
                        ((tk.agent.agent_idx - 1) % village_size, (tk.agent.agent_idx - 1) % village_size)
                    )
                # VOTE
                if content.topic == Topic.VOTE:
                    if self.vote_declare[(tk.agent.agent_idx - 1) % village_size] != -1:
                        self.update_queue.append(
                            (
                                (tk.agent.agent_idx - 1) % village_size,
                                self.vote_declare[(tk.agent.agent_idx - 1) % village_size]
                            )
                        )
                    self.vote_declare[
                        (tk.agent.agent_idx - 1) % village_size
                        ] = (content.target.agent_idx - 1) % village_size
                    self.pair_features[(tk.agent.agent_idx - 1) % village_size, :, 1] = 0
                    self.pair_features[
                        (tk.agent.agent_idx - 1) % village_size, (content.target.agent_idx - 1) % village_size, 1
                    ] = 1
                    self.update_queue.append(
                        (
                            (tk.agent.agent_idx - 1) % village_size,
                            self.vote_declare[(tk.agent.agent_idx - 1) % village_size]
                        )
                    )

                # DIVINED
                if content.topic == Topic.DIVINED:
                    if content.result == Species.WEREWOLF:
                        self.pair_features[
                            (tk.agent.agent_idx - 1) % village_size, (content.target.agent_idx - 1) % village_size, 2
                        ] = 1
                        self.update_queue.append(
                            (
                                (tk.agent.agent_idx - 1) % village_size,
                                (content.target.agent_idx - 1) % village_size
                            )
                        )
                    elif content.result == Species.HUMAN:
                        self.pair_features[
                            (tk.agent.agent_idx - 1) % village_size, (content.target.agent_idx - 1) % village_size, 3
                        ] = 1
                        self.update_queue.append(
                            (
                                (tk.agent.agent_idx - 1) % village_size,
                                (content.target.agent_idx - 1) % village_size
                            )
                        )

                # IDENTIFIED
                if content.topic == Topic.IDENTIFIED:
                    if content.result == Species.WEREWOLF:
                        self.pair_features[
                            (tk.agent.agent_idx - 1) % village_size, (content.target.agent_idx - 1) % village_size, 4
                        ] = 1
                        self.update_queue.append(
                            (
                                (tk.agent.agent_idx - 1) % village_size,
                                (content.target.agent_idx - 1) % village_size
                            )
                        )
                    elif content.result == Species.HUMAN:
                        self.pair_features[
                            (tk.agent.agent_idx - 1) % village_size, (content.target.agent_idx - 1) % village_size, 5
                        ] = 1
                        self.update_queue.append(
                            (
                                (tk.agent.agent_idx - 1) % village_size,
                                (content.target.agent_idx - 1) % village_size
                            )
                        )

                # GUARDED
                if content.topic == Topic.GUARDED:
                    self.pair_features[
                        (tk.agent.agent_idx - 1) % village_size, (content.target.agent_idx - 1) % village_size, 6
                    ] = 1
                    self.update_queue.append(
                        (
                            (tk.agent.agent_idx - 1) % village_size,
                            (content.target.agent_idx - 1) % village_size
                        )
                    )

                self.talk_idx += 1

        # read vote if re-vote happens
        for vote in game_info.latest_vote_list:
            self.pair_features[
                (vote.agent.agent_idx - 1) % village_size, (vote.target.agent_idx - 1) % village_size, 0
            ] = 1
            self.update_queue.append(
                ((vote.agent.agent_idx - 1) % village_size, (vote.target.agent_idx - 1) % village_size)
            )

        # print(len(self.update_queue))
        while len(self.update_queue):
            i, j = self.update_queue.pop()
            if i == j:
                new_odds = self.get_single_log_odds(i)
                # print(i, j, new_odds)
                v, p, w = new_odds
                self.log_odds_list[case_mat[:, i] == 0] += v - self.single_log_odds[i, 0]
                self.log_odds_list[case_mat[:, i] == 1] += p - self.single_log_odds[i, 1]
                self.log_odds_list[case_mat[:, i] == 2] += w - self.single_log_odds[i, 2]
                # new -> old
                self.single_log_odds[i, :] = new_odds
            else:
                new_odds = self.get_pair_log_odds(i, j)
                # print(i, j, new_odds)
                vv, vp, vw, pv, pw, wv, wp, ww = new_odds
                self.log_odds_list[mat_vv[i, j] == 1] += vv - self.pair_log_odds[i, j, 0]
                self.log_odds_list[mat_vp[i, j] == 1] += vp - self.pair_log_odds[i, j, 1]
                self.log_odds_list[mat_vw[i, j] == 1] += vw - self.pair_log_odds[i, j, 2]
                self.log_odds_list[mat_pv[i, j] == 1] += pv - self.pair_log_odds[i, j, 3]
                self.log_odds_list[mat_pw[i, j] == 1] += pw - self.pair_log_odds[i, j, 4]
                self.log_odds_list[mat_wv[i, j] == 1] += wv - self.pair_log_odds[i, j, 5]
                self.log_odds_list[mat_wp[i, j] == 1] += wp - self.pair_log_odds[i, j, 6]
                self.log_odds_list[mat_ww[i, j] == 1] += ww - self.pair_log_odds[i, j, 7]
                # new -> old
                self.pair_log_odds[i, j, :] = new_odds
        # print(time.time() - t0)
        return None

    def get_single_log_odds(self, i):
        res_log_odds = np.zeros(3, dtype=np.float32)
        # attacked
        if self.single_features[i, 7] > 0:
            # werewolves never attacked
            res_log_odds[2] -= 100.0
        # co seer
        if self.single_features[i, 1] > 0:
            if self.day <= 1:
                res_log_odds += self.agent_param[i, -9:-6, 0]
            else:
                res_log_odds += self.agent_param[i, -9:-6, 1]
        # co medium
        elif self.single_features[i, 2] > 0:
            if self.day <= 1:
                res_log_odds += self.agent_param[i, -6:-3, 0]
            else:
                res_log_odds += self.agent_param[i, -6:-3, 2]
        # co bodyguard
        elif self.single_features[i, 3] > 0:
            if self.day <= 1:
                res_log_odds += self.agent_param[i, -3:, 0]
            else:
                res_log_odds += self.agent_param[i, -3:, 1]
        return res_log_odds

    def get_pair_log_odds(self, i, j):
        res_log_odds = np.zeros(8, dtype=np.float32)
        if self.day <= 1:
            vec = self.agent_param[i, :8 * self.n_pair_features, 0]
        else:
            vec = self.agent_param[i, :8 * self.n_pair_features, 1]
        for k in range(self.n_pair_features):
            if self.pair_features[i, j, k] > 0:
                res_log_odds += vec[(k * 8):((k + 1) * 8)]
        return res_log_odds

    def get_prob(self, wn=0, sp=0):
        # count
        cnt_alive_humans = np.zeros(case_size, dtype=np.int32)
        cnt_alive_wolves = np.zeros(case_size, dtype=np.int32)
        for i in range(village_size):
            if self.status[i] == 0:
                cnt_alive_humans += (case_mat[:, i] < 2).astype(np.int32)
                cnt_alive_wolves += (case_mat[:, i] == 2).astype(np.int32)

        cnt_village_seers = np.zeros(case_size, dtype=np.int32)
        cnt_village_mediums = np.zeros(case_size, dtype=np.int32)
        cnt_village_bodyguards = np.zeros(case_size, dtype=np.int32)
        cnt_enemy_seers = np.zeros(case_size, dtype=np.int32)
        cnt_enemy_mediums = np.zeros(case_size, dtype=np.int32)
        cnt_enemy_bodyguards = np.zeros(case_size, dtype=np.int32)
        for i in range(village_size):
            if self.co[i] == 1:
                cnt_village_seers += (case_mat[:, i] == 0).astype(np.int32)
                cnt_enemy_seers += (case_mat[:, i] != 0).astype(np.int32)
            elif self.co[i] == 2:
                cnt_village_mediums += (case_mat[:, i] == 0).astype(np.int32)
                cnt_enemy_mediums += (case_mat[:, i] != 0).astype(np.int32)
            elif self.co[i] == 3:
                cnt_village_bodyguards += (case_mat[:, i] == 0).astype(np.int32)
                cnt_enemy_bodyguards += (case_mat[:, i] != 0).astype(np.int32)

        log_odds_list = self.log_odds_list.copy()
        if self.day <= 1:
            # villagers should not co
            log_odds_list[cnt_village_seers >= 2] -= 2.5
            log_odds_list[cnt_village_mediums >= 2] -= 1.8
            log_odds_list[cnt_village_bodyguards >= 2] -= 0

            # should co
            log_odds_list[(cnt_village_seers == 0) * (cnt_enemy_seers > 0)] -= 0.3
            log_odds_list[(cnt_village_mediums == 0) * (cnt_enemy_mediums > 0)] -= 0.06
            log_odds_list[(cnt_village_bodyguards == 0) * (cnt_enemy_bodyguards > 0)] += 0.13
        else:
            # villagers should not co
            log_odds_list[cnt_village_seers >= 2] -= 2.2
            log_odds_list[cnt_village_mediums >= 2] -= 1.6
            log_odds_list[cnt_village_bodyguards >= 2] -= 0.44

            # should co
            log_odds_list[(cnt_village_seers == 0) * (cnt_enemy_seers > 0)] -= 0.47
            log_odds_list[(cnt_village_mediums == 0) * (cnt_enemy_mediums > 0)] -= 0.39
            log_odds_list[(cnt_village_bodyguards == 0) * (cnt_enemy_bodyguards > 0)] += 0.11

        res_prob = np.zeros((village_size, 3), dtype=np.float32)

        prob_list = np.exp(log_odds_list)
        prob_list = prob_list / (1 + prob_list)

        # game goes on
        prob_list[cnt_alive_humans <= cnt_alive_wolves] *= 0.0
        prob_list[cnt_alive_wolves == 0] *= 0.0

        # WATASHI NINGEN!
        if 1 <= wn <= village_size:
            prob_list[case_mat[:, wn - 1] != sp] *= 0.0

        # standardize
        prob_list /= prob_list.sum()

        for i in range(village_size):
            res_prob[i, 0] = prob_list[case_mat[:, i] == 0].sum()
            res_prob[i, 1] = prob_list[case_mat[:, i] == 1].sum()
            res_prob[i, 2] = prob_list[case_mat[:, i] == 2].sum()
        return res_prob


if __name__ == "__main__":
    re = RoleEstimator05()

