from typing import Optional, List
from collections import deque
import random

import numpy as np

from aiwolf import (Agent, ComingoutContentBuilder, Content, GameInfo, GameSetting,
                    DivinedResultContentBuilder, IdentContentBuilder, Judge, Role, Species, Status,
                    VoteContentBuilder, AttackContentBuilder, SkipContentBuilder, OverContentBuilder)
from aiwolf.gameinfo import GameInfo, _GameInfo
from aiwolf.gamesetting import GameSetting, _GameSetting
from aiwolf.utterance import Talk, Whisper, _Utterance
from aiwolf.constant import AGENT_NONE
from aiwolfpy.player import AbstractPlayer
from aiwolfpy.parser import Parser
from aiwolfpy.model import RoleEstimator15, RoleEstimator05, AgentEstimator15, AgentEstimator05


class Cash(AbstractPlayer):

    def __init__(self):
        self.parser = Parser()
        self.me = None
        self.role = Role.UNC
        self.co = Role.UNC
        self.werewolves = []
        self.humans = []
        self.talk_list = []
        self.my_judge_queue = deque()
        self.divined = []
        self.vote_target = AGENT_NONE
        self.vote_declare = AGENT_NONE
        self.attack_vote_target = AGENT_NONE
        self.attack_vote_declare = AGENT_NONE
        self.divine_target = AGENT_NONE
        self.guard_target = AGENT_NONE
        self.role_estimator05 = RoleEstimator05()
        self.role_estimator15 = RoleEstimator15()
        self.agent_estimator05 = AgentEstimator05()
        self.agent_estimator15 = AgentEstimator15()
        self.role_estimator = None
        self.agent_estimator = None
        self.village_size = -1
        self.win_rate_ww = None
        self.fake_identify = []

    def initialize(self, game_info: _GameInfo, game_setting: _GameSetting):
        # initialize parser
        self.parser.initialize(game_info, game_setting)
        game_info = self.parser.game_info

        # initialize estimators
        if self.village_size != game_setting["playerNum"]:
            if game_setting["playerNum"] == 5:
                self.role_estimator = self.role_estimator05
                self.agent_estimator = self.agent_estimator05
            else:
                self.role_estimator = self.role_estimator15
                self.agent_estimator = self.agent_estimator15
            self.village_size = game_setting["playerNum"]
            self.win_rate_ww = np.zeros((self.village_size, 2))
        else:
            self.role_estimator.initialize(self.agent_estimator.agent_prob)

        # my role
        self.me = game_info.me
        self.role = game_info.my_role
        self.co = Role.UNC
        self.werewolves = []
        self.humans = []
        self.talk_list = []
        self.my_judge_queue.clear()
        self.divined = []
        self.vote_target = self.me
        self.vote_declare = AGENT_NONE
        self.attack_vote_target = self.me
        self.attack_vote_declare = AGENT_NONE
        self.divine_target = self.me
        self.guard_target = self.me
        self.fake_identify = [0] * self.village_size

        # for werewolves
        if self.role == Role.WEREWOLF:
            self.werewolves = list(game_info.role_map.keys())
            self.humans = [a for a in game_info.agent_list if a not in self.werewolves]

    def update(
            self, game_info: _GameInfo, talk_history: Optional[List[_Utterance]],
            whisper_history: Optional[List[_Utterance]], request
    ):
        # print(request)
        # print(game_info)
        # print(talk_history)
        # print(whisper_history)
        self.parser.update(game_info, talk_history, whisper_history, request)
        if request != "FINISH":
            self.role_estimator.update(self.parser.game_info)

        if len(self.talk_list):
            last_talk = self.talk_list[-1]
            last_talk_idx = last_talk.idx
            last_talk_day = last_talk.day
        else:
            last_talk_idx = -1
            last_talk_day = -1
        for talk in self.parser.game_info.talk_list:
            if talk.idx > last_talk_idx or talk.day > last_talk_day:
                self.talk_list.append(talk)

    def day_start(self):
        self.vote_target = self.me
        self.vote_declare = AGENT_NONE
        self.attack_vote_target = self.me
        self.attack_vote_declare = AGENT_NONE
        self.divine_target = self.me
        self.guard_target = self.me

        # get result
        # SEER
        if self.role == Role.SEER:
            judge: Optional[Judge] = self.parser.game_info.divine_result
            if judge is not None:
                self.my_judge_queue.append(judge)
        # MEDIUM
        if self.role == Role.MEDIUM:
            judge: Optional[Judge] = self.parser.game_info.medium_result
            if judge is not None:
                self.my_judge_queue.append(judge)
        # FAKE SEER RESULT
        if self.role != self.co == Role.SEER:
            # random target
            candidate_list = []
            for a in self.parser.game_info.agent_list:
                if a == self.me:
                    continue
                if self.parser.game_info.status_map[a] == Status.ALIVE:
                    candidate_list.append(a)
            for a in self.parser.game_info.last_dead_agent_list:
                candidate_list.append(a)
            target: Agent = random.choice(candidate_list)
            # random result
            result: Species = Species.HUMAN
            d = 0.5
            if random.random() < d and target not in self.parser.game_info.last_dead_agent_list:
                result = Species.WEREWOLF
            judge = Judge(self.me, self.parser.game_info.day, target, result)
            self.my_judge_queue.append(judge)

        return None

    def talk(self):
        alive_cnt = 0
        for a in self.parser.game_info.status_map.keys():
            st = self.parser.game_info.status_map[a]
            if st.name == "ALIVE":
                alive_cnt += 1
        co_ps = 0
        if self.village_size == 5:
            for i, v in enumerate(self.role_estimator.co):
                if v == 2:
                    co_ps += 1
        else:
            for i, v in enumerate(self.role_estimator.co):
                if v == 5:
                    co_ps += 1
        # co anyway
        if self.co == Role.UNC:
            if self.role == Role.SEER:
                self.co = Role.SEER
                content = Content(ComingoutContentBuilder(target=self.me, role=Role.SEER))
                return content.text
            elif self.role == Role.MEDIUM:
                self.co = Role.MEDIUM
                content = Content(ComingoutContentBuilder(target=self.me, role=Role.MEDIUM))
                return content.text
            elif self.role == Role.POSSESSED:
                # co as seer
                self.co = Role.SEER
                content = Content(ComingoutContentBuilder(target=self.me, role=Role.SEER))
                # add fake result for day 1
                # 人狼での勝率が１番低い人に黒出し
                if self.village_size == 15:
                    wr_ww = (2 + self.win_rate_ww[:, 1]) / (10 + self.win_rate_ww[:, 0])
                else:
                    wr_ww = (4 + self.win_rate_ww[:, 1]) / (10 + self.win_rate_ww[:, 0])
                # 自分は除く
                wr_ww[self.me.agent_idx - 1] = 1.0
                target: Agent = self.parser.game_info.agent_list[np.argmin(wr_ww)]
                # random result
                result = Species.WEREWOLF
                judge = Judge(self.me, self.parser.game_info.day, target, result)
                self.my_judge_queue.append(judge)
                return content.text
            else:
                self.co = Role.VILLAGER
        # PP
        elif self.role == Role.POSSESSED and self.co != Role.POSSESSED:
            # 狂人の場合は残り3人で狂人CO
            if alive_cnt == 3:
                self.co = Role.POSSESSED
                content = Content(ComingoutContentBuilder(target=self.me, role=Role.POSSESSED))
                return content.text
        elif co_ps >= 1 and alive_cnt == 3 and self.co != Role.POSSESSED and self.co != Role.WEREWOLF:
            # それ以外は狂人COを見つけたら人狼CO
            self.co = Role.WEREWOLF
            content = Content(ComingoutContentBuilder(target=self.me, role=Role.WEREWOLF))
            return content.text
        # what about bodyguard? powerplay?

        # result
        if len(self.my_judge_queue) > 0:
            if self.role == self.co == Role.SEER:
                judge: Judge = self.my_judge_queue.popleft()
                content = Content(DivinedResultContentBuilder(judge.target, judge.result))
                return content.text
            elif self.role == self.co == Role.MEDIUM:
                judge: Judge = self.my_judge_queue.popleft()
                content = Content(IdentContentBuilder(judge.target, judge.result))
                return content.text
            elif self.co == Role.SEER:
                judge: Judge = self.my_judge_queue.popleft()
                content = Content(DivinedResultContentBuilder(judge.target, judge.result))
                return content.text

        # think who to vote for
        self._think_vote()

        # fake_identify
        if max(self.fake_identify) >= 10000:
            for i in range(self.village_size):
                if self.fake_identify[i] >= 10000:
                    self.fake_identify[i] = 1
                    return f"IDENTIFIED Agent[{str(101 + i)[1:]}] WEREWOLF"

        # declare who to vote for
        if self.vote_declare is AGENT_NONE or self.vote_declare != self.vote_target:
            self.vote_declare = self.vote_target
            content = Content(VoteContentBuilder(target=self.vote_declare))
            return content.text

        # something to talk later
        # if False:
        #     content = Content(SkipContentBuilder())
        #     return content.text

        # nothing to talk anymore
        content = Content(OverContentBuilder())
        return content.text

    def whisper(self):
        if self.co == Role.UNC:
            self.co = Role.VILLAGER
            content = Content(ComingoutContentBuilder(target=self.me, role=Role.VILLAGER))
            return content.text

        if self.parser.game_info.day > 0:
            # think who to attack
            # think once because it is a sample
            if self.attack_vote_declare == AGENT_NONE:
                self._think_attack_vote()
            # declare who to attack
            if self.attack_vote_declare is AGENT_NONE or self.attack_vote_declare != self.attack_vote_target:
                self.attack_vote_declare = self.attack_vote_target
                content = Content(AttackContentBuilder(target=self.attack_vote_declare))
                return content.text
        # something to talk later
        # if False:
        #     content = Content(SkipContentBuilder())
        #     return content.text

        # nothing to talk anymore
        content = Content(OverContentBuilder())
        return content.text

    def vote(self):
        self._think_vote()
        # print(self.role_estimator.get_prob())
        return self.vote_target.agent_idx

    def attack(self):
        return self.attack_vote_target.agent_idx

    def divine(self):
        # divine most likely werewolf
        p_mat = self.role_estimator.get_prob(wn=self.me.agent_idx)
        p_vote = p_mat[:, 2]

        # No Vote for dead agents
        for a in self.parser.game_info.agent_list:
            if self.parser.game_info.status_map[a] != Status.ALIVE:
                p_vote[a.agent_idx - 1] *= 0.0
        # No vote for me
        p_vote[self.me.agent_idx - 1] *= 0.0

        # no divine again
        for idx in self.divined:
            p_vote[idx - 1] *= 0.00001

        p_max = p_vote.max()
        candidate_list = []
        for i, a in enumerate(self.parser.game_info.agent_list):
            if p_vote[i] >= p_max - 0.000000001:
                candidate_list.append(a)
        self.divine_target = random.choice(candidate_list)
        self.divined.append(self.divine_target.agent_idx)
        return self.divine_target.agent_idx

    def guard(self):
        # guard most likely villager
        p_mat = self.role_estimator.get_prob(wn=self.me.agent_idx)
        # guard if seer 1CO or medium 1CO
        # COをカウントする
        seer_cnt = 0
        medium_cnt = 0
        for i, v in enumerate(self.role_estimator.co):
            if v == 1:
                seer_cnt += 1
            elif v == 2:
                medium_cnt += 1
        p_vote = p_mat[:, 0]
        for i, v in enumerate(self.role_estimator.co):
            if v == 1 and seer_cnt == 1:
                p_vote[i] += 1.0
            elif v == 2 and medium_cnt == 1:
                p_vote[i] += 0.7
        # No Vote for dead agents
        for a in self.parser.game_info.agent_list:
            if self.parser.game_info.status_map[a] != Status.ALIVE:
                p_vote[a.agent_idx - 1] *= 0.0
        # No vote for me
        p_vote[self.me.agent_idx - 1] *= 0.0

        p_max = p_vote.max()
        candidate_list = []
        for i, a in enumerate(self.parser.game_info.agent_list):
            if p_vote[i] >= p_max - 0.000000001:
                candidate_list.append(a)
        self.guard_target = random.choice(candidate_list)
        return self.guard_target.agent_idx

    def finish(self):
        # you can update your AI if you want and can
        self.agent_estimator.update(self.parser.game_info.role_map, self.talk_list)
        # print([(talk.day, talk.idx, talk.agent.agent_idx, talk.text) for talk in self.talk_list])
        # for k in self.parser.game_info.role_map:
        #     v = self.parser.game_info.role_map[k]
        #     print(k.agent_idx, v.name)

        # 人狼の勝率をupdate
        w = 0.0
        for k in self.parser.game_info.role_map.keys():
            v = self.parser.game_info.role_map[k]
            if v.name == "WEREWOLF":
                st = self.parser.game_info.status_map[k]
                if st.name == "ALIVE":
                    w = 1.0
        for k in self.parser.game_info.role_map.keys():
            v = self.parser.game_info.role_map[k]
            if v.name == "WEREWOLF":
                self.win_rate_ww[k.agent_idx - 1, 0] += 1.0
                self.win_rate_ww[k.agent_idx - 1, 1] += w
        return None

    def _think_vote(self):
        # 村用
        p_mat = self.role_estimator.get_prob(wn=self.me.agent_idx)
        p_vote = p_mat[:, 2] + 0.5 * p_mat[:, 1]
        # COをカウントする
        seer_cnt = 0
        medium_cnt = 0
        guard_cnt = 0
        for i, v in enumerate(self.role_estimator.co):
            if v == 1:
                seer_cnt += 1
            elif v == 2:
                medium_cnt += 1
            elif v == 3:
                guard_cnt += 1
        # 1COは基本吊らない
        if self.village_size == 15:
            for i, v in enumerate(self.role_estimator.co):
                if seer_cnt == 1 and v == 1:
                    p_vote[i] *= 0.5
                elif medium_cnt == 1 and v == 2:
                    p_vote[i] *= 0.5
                elif guard_cnt == 1 and v == 3:
                    p_vote[i] *= 0.7

        # 初日から占いは吊らない（3-1は除く）
        if self.parser.game_info.day <= 1 and self.village_size == 15:
            if seer_cnt <= 2 or medium_cnt >= 2:
                for i, v in enumerate(self.role_estimator.co):
                    if v == 1:
                        p_vote[i] *= 0.1

        if self.role == Role.WEREWOLF:
            # 身内切りは避ける
            for a in self.werewolves:
                p_vote[a.agent_idx - 1] *= 0.5
            # 推定狂人も避ける
            p_mat_ww = self.role_estimator.get_prob(wn=self.me.agent_idx, sp=2)
            p_vote -= p_mat_ww[:, 1] * 0.5
            p_vote = np.maximum(p_vote, 0.000001)

        elif self.role == Role.POSSESSED:
            # 推定人狼には投票しない
            p_mat_ps = self.role_estimator.get_prob(wn=self.me.agent_idx, sp=1)
            p_vote -= p_mat_ps[:, 2] * 0.9
            p_vote = np.maximum(p_vote, 0.000001)

            # PP
            if self.co == Role.POSSESSED:
                p_mat_ps = self.role_estimator.get_prob(wn=self.me.agent_idx, sp=1)
                p_vote = p_mat_ps[:, 0]
                # 人狼COが一人ならその人には投票しない
                ww_cnt = 0
                ww_co = -1
                if self.village_size == 5:
                    for i, v in enumerate(self.role_estimator.co):
                        if v == 3:
                            ww_cnt += 1
                            ww_co = i
                else:
                    for i, v in enumerate(self.role_estimator.co):
                        if v == 6:
                            ww_cnt += 1
                            ww_co = i
                if ww_cnt == 1:
                    p_vote[ww_co] *= 0.0001

        # PP or PP返し
        if self.co == Role.WEREWOLF:
            if self.village_size == 5:
                for i, v in enumerate(self.role_estimator.co):
                    if v == 2:
                        p_vote[i] *= 0.001
            else:
                for i, v in enumerate(self.role_estimator.co):
                    if v == 5:
                        p_vote[i] *= 0.001

        # No Vote for dead agents
        for a in self.parser.game_info.agent_list:
            if self.parser.game_info.status_map[a] != Status.ALIVE:
                p_vote[a.agent_idx - 1] *= 0.0
        # No vote for me
        p_vote[self.me.agent_idx - 1] *= 0.0

        p_max = p_vote.max()
        candidate_list = []
        for i, a in enumerate(self.parser.game_info.agent_list):
            if p_vote[i] >= p_max - 0.000000001:
                candidate_list.append(a)
        if self.vote_target not in candidate_list:
            self.vote_target = random.choice(candidate_list)

        # 人狼の場合用
        if self.role == Role.WEREWOLF and self.co != Role.WEREWOLF and p_max > 0.4:
            if self.fake_identify[self.vote_target.agent_idx - 1] == 0:
                self.fake_identify[self.vote_target.agent_idx - 1] += 10000

        return None

    def _think_attack_vote(self):
        p_mat = self.role_estimator.get_prob(wn=self.me.agent_idx, sp=2)
        # COをカウントする
        seer_cnt = 0
        medium_cnt = 0
        for i, v in enumerate(self.role_estimator.co):
            if v == 1:
                seer_cnt += 1
            elif v == 2:
                medium_cnt += 1
        p_vote = p_mat[:, 0]
        for i, v in enumerate(self.role_estimator.co):
            if v == 1 and seer_cnt > 1:
                p_vote[i] += 0.5
            elif v == 2 and medium_cnt > 1:
                p_vote[i] += 0.2

        # No Vote for dead agents
        for a in self.parser.game_info.agent_list:
            if self.parser.game_info.status_map[a] != Status.ALIVE:
                p_vote[a.agent_idx - 1] *= 0.0
        # No vote for me
        p_vote[self.me.agent_idx - 1] *= 0.0
        # No vote for werewolves
        for a in self.werewolves:
            p_vote[a.agent_idx - 1] *= 0.0

        p_max = p_vote.max()
        candidate_list = []
        for i, a in enumerate(self.parser.game_info.agent_list):
            if p_vote[i] >= p_max - 0.000000001:
                candidate_list.append(a)
        self.attack_vote_target = random.choice(candidate_list)
        return None
